package org.frogpond.service;

import org.frogpond.LilyException;
import org.frogpond.annotation.LilyField;
import org.frogpond.metadata.RecordFieldMetadata;
import org.frogpond.metadata.RecordTypeMetadata;
import org.frogpond.metadata.store.MetadataStore;
import org.frogpond.model.Primitive;
import org.frogpond.model.PropertyEntry;
import org.frogpond.utils.LilyUtilities;
import org.frogpond.utils.MetadataUtilities;
import org.lilyproject.repository.api.*;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

public class LilyRecordConverter {
    private MetadataStore metadataStore;
    private Repository repository;

    public LilyRecordConverter() {
    }

    public LilyRecordConverter(MetadataStore metadataStore, Repository repository) {
        this.metadataStore = metadataStore;
        this.repository = repository;
    }

    public void copyIntoPojo(Record record, RecordTypeMetadata recordType, Object pojo) throws LilyException {
        for (RecordFieldMetadata recordFieldMetadata : recordType.getFields()) {
            QName fieldName = recordFieldMetadata.getFieldType().getFieldType().getName();

            try {
                // -- check if the record has the current field
                if (! record.hasField(fieldName)) continue;

                // -- get the record value
                Object recordValue = record.getField(fieldName);

                // -- set the field value
                recordFieldMetadata.getAccessor().setValue(pojo, recordValue);
            } catch (Exception e) {
                throw new LilyException(String.format(
                        "Unable to copy data into field %s of record %s",
                        recordFieldMetadata.getFieldType().getFieldType().getName(),
                        recordFieldMetadata.getRecordType().getRecordType().getName()
                ), e);
            }
        }

        // -- we need to copy the recordId into the field annotated with @LilyId in case the field isn't annotated
        // -- with @LilyField too.
        PropertyEntry recordIdProperty = recordType.getRecordIdProperty();
        if (recordIdProperty.getAnnotation(LilyField.class) == null) {
            try {
                if (RecordId.class.isAssignableFrom(recordIdProperty.getType())) {
                    // -- set the field value
                    MetadataUtilities.getFieldAccessor(recordIdProperty).setValue(pojo, record.getId());
                } else if (String.class.isAssignableFrom(recordIdProperty.getType())) {
                    // -- set the field value
                    MetadataUtilities.getFieldAccessor(recordIdProperty).setValue(pojo, record.getId().toString());
                } else {
                    throw new Exception("The lily record id cannot be converted to a " + recordIdProperty.getType().getName());
                }
            } catch (Exception e) {
                throw new LilyException(String.format(
                        "Unable to copy the system generated id into field %s of record %s",
                        recordIdProperty.getName(),
                        recordType.getName()
                ), e);
            }
        }
    }

    public void copyIntoRecord(Record record, RecordTypeMetadata recordType, Object pojo) throws LilyException {
        record.setRecordType(recordType.getName(), recordType.getVersion());

        for (RecordFieldMetadata recordFieldMetadata : recordType.getFields()) {
            try {
                Object pojoValue = recordFieldMetadata.getAccessor().getValue(pojo);

                // -- in case of a link field, we need to construct a link
                if ((pojoValue != null) && (recordFieldMetadata.getPrimitive() == Primitive.Link)) {
                    if (recordFieldMetadata.isMultiValue()) {
                        // -- get the lilyField annotation
                        LilyField fieldAnnotation = recordFieldMetadata.getFieldType().getProperty().getAnnotation(LilyField.class);

                        // -- check if the field annotation holds a java type
                        if (fieldAnnotation.javaType() == Object.class)
                            throw new LilyException("A Java type should be declared when a field holds a generic collection.");

                        // -- get the recordTypeMetadata for the type
                        RecordTypeMetadata linkRecordTypeMetadata = metadataStore.getRecordTypeMetadata(
                                fieldAnnotation.javaType()
                        );

                        pojoValue = createLinkCollection(linkRecordTypeMetadata, (Collection<?>) pojoValue);
                    } else {
                        // -- get the recordTypeMetadata for the type
                        RecordTypeMetadata linkRecordTypeMetadata = metadataStore.getRecordTypeMetadata(
                                recordFieldMetadata.getFieldType().getProperty().getType()
                        );

                        pojoValue = createLink(linkRecordTypeMetadata, pojoValue);
                    }
                }

                if (pojoValue == null) continue;

                // -- set the record value
                record.setField(recordFieldMetadata.getFieldType().getFieldType().getName(), pojoValue);
            } catch (Exception e) {
                throw new LilyException(String.format(
                        "Unable to copy data from field %s of record %s",
                        recordFieldMetadata.getFieldType().getFieldType().getName(),
                        recordFieldMetadata.getRecordType().getRecordType().getName()
                ), e);
            }
        }
    }

    protected List<Link> createLinkCollection(RecordTypeMetadata recordTypeMetadata, Collection<?> collection) throws LilyException {
        List<Link> links = new ArrayList<Link>();

        for (Object collectionItem : collection) {
            links.add(createLink(recordTypeMetadata, collectionItem));
        }

        return links;
    }

    protected Link createLink(RecordTypeMetadata recordTypeMetadata, Object fieldValue) throws LilyException {
        if (fieldValue instanceof Link) {
            return ((Link) fieldValue);
        } else {
            // -- get the recordId of the pojo value
            RecordId recordId = LilyUtilities.getRecordId(repository, recordTypeMetadata, fieldValue);

            // -- construct the link
            return new Link(recordId);
        }
    }

    public MetadataStore getMetadataStore() {
        return metadataStore;
    }

    public void setMetadataStore(MetadataStore metadataStore) {
        this.metadataStore = metadataStore;
    }

    public Repository getRepository() {
        return repository;
    }

    public void setRepository(Repository repository) {
        this.repository = repository;
    }
}
