package org.frogpond.service.resolvers;

import org.frogpond.LilyException;
import org.frogpond.metadata.RecordFieldMetadata;
import org.frogpond.metadata.RecordTypeMetadata;
import org.frogpond.metadata.store.MetadataStore;
import org.frogpond.model.Primitive;
import org.frogpond.service.LilyRecordConverter;
import org.lilyproject.repository.api.*;
import org.springframework.core.CollectionFactory;

import java.util.Collection;

public class LinkFieldResolver implements FieldResolver {
    private Repository repository;
    private MetadataStore metadataStore;
    private LilyRecordConverter recordConverter;

    @Override
    public Primitive getPrimitive() {
        return Primitive.Link;
    }

    @Override
    public Object resolve(RecordFieldMetadata recordField, Record record) {
        // -- get the field
        Object obj = record.getField(recordField.getQName());

        if (obj instanceof Collection) {
            Collection collection = CollectionFactory.createCollection(recordField.getFieldType().getProperty().getType(), ((Collection) obj).size());

            for (Object item : (Collection) obj) {
                if (item instanceof Link)
                    collection.add(resolveLink(recordField, record, (Link) item));
            }

            return collection;
        } else {
            return resolveLink(recordField, record, (Link) obj);
        }
    }

    public Object resolveLink(RecordFieldMetadata recordField, Record record, Link link) {
        // -- check if a link is set
        if (link == null) return null;

        // -- resolve the link
        RecordId linkedRecordId = link.resolve(record, repository.getIdGenerator());

        if (linkedRecordId == null) return null;

        try {
            // -- get the linked record
            Record linkedRecord = repository.read(linkedRecordId);

            // -- get the metadata for the linked record type
            QName linkedRecordTypeQName = linkedRecord.getRecordTypeName(Scope.VERSIONED);
            RecordTypeMetadata linkedRecordTypeMetadata = metadataStore.getRecordTypeMetadata(linkedRecordTypeQName);

            // -- check if the metadata was available
            if (linkedRecordTypeMetadata == null)
                throw new LilyException("No recordType found for " + linkedRecord.getRecordTypeName());

            // -- convert the linked record into an object
            Object result = linkedRecordTypeMetadata.getJavaType().newInstance();
            recordConverter.copyIntoPojo(linkedRecord, linkedRecordTypeMetadata, result);

            // -- return the resulting object
            return result;
        } catch (Exception e) {
            return null;
        }
    }

    public void setRepository(Repository repository) {
        this.repository = repository;
    }

    public void setRecordConverter(LilyRecordConverter recordConverter) {
        this.recordConverter = recordConverter;
    }

    public void setMetadataStore(MetadataStore metadataStore) {
        this.metadataStore = metadataStore;
    }
}
