package org.frogpond.service;

import org.apache.log4j.Logger;
import org.frogpond.LilyException;
import org.frogpond.demo.model.Author;
import org.frogpond.demo.model.Book;
import org.frogpond.metadata.extractor.ReflectiveMetadataExtractor;
import org.frogpond.metadata.store.DefaultMetadataStoreFactory;
import org.frogpond.metadata.store.MetadataStore;
import org.frogpond.model.Primitive;
import org.frogpond.service.resolvers.DefaultFieldResolver;
import org.frogpond.service.resolvers.DefaultFieldResolverManager;
import org.frogpond.service.resolvers.FieldResolver;
import org.frogpond.service.resolvers.LinkFieldResolver;
import org.lilyproject.client.LilyClient;
import org.lilyproject.repository.api.Repository;

import java.util.Arrays;
import java.util.Collections;

public class BasicLilyServiceDemo {
    private static final Logger LOGGER = Logger.getLogger(BasicLilyServiceDemo.class);
    private BasicLilyService lilyService;

    public static void main(String[] args) {
        BasicLilyServiceDemo demo = new BasicLilyServiceDemo();

        try {
            // -- perform initialization
            demo.initialize(args[0]);

            demo.demoCreate();
            demo.demoRetrieval();

            //demo.demoResolve();
            demo.demoUpdate();
            //demo.demoDelete();

        } catch (Exception e) {
            LOGGER.error("Unable to run the demo!", e);
        }
    }

    protected void initialize(String zookeeperUrl) throws Exception {
        // -- create the lily client
        LilyClient lilyClient = new LilyClient(zookeeperUrl, 60000);
        Repository repository = lilyClient.getRepository();

        // -- construct the metadata store
        DefaultMetadataStoreFactory metadataStoreFactory = new DefaultMetadataStoreFactory();
        metadataStoreFactory.setMetadataExtractor(new ReflectiveMetadataExtractor());
        metadataStoreFactory.setBasePackage("org.frogpond.demo.model");
        MetadataStore metadataStore = metadataStoreFactory.create();

        // -- create the record converter
        LilyRecordConverter recordConverter = new LilyRecordConverter(metadataStore, repository);

        // -- create the field resolver
        DefaultFieldResolverManager fieldResolverManager = new DefaultFieldResolverManager();
        fieldResolverManager.setDefaultResolver(new DefaultFieldResolver());
        fieldResolverManager.setResolvers(Collections.singletonMap(Primitive.Link, createLinkFieldResolver(repository, metadataStore, recordConverter)));

        // -- create the lily service
        lilyService = new BasicLilyService();
        lilyService.setRepository(repository);
        lilyService.setFieldResolverManager(fieldResolverManager);
        lilyService.setMetadataStore(metadataStore);
        lilyService.setRecordConverter(recordConverter);
    }

    private FieldResolver createLinkFieldResolver(Repository repository, MetadataStore metadataStore, LilyRecordConverter recordConverter) {
        LinkFieldResolver result = new LinkFieldResolver();
        result.setRecordConverter(recordConverter);
        result.setRepository(repository);
        result.setMetadataStore(metadataStore);
        return result;
    }

    public void demoCreate() {
        Author daan = new Author("Daan Gerits", "A developer with his mind in the cloud's.", "daan.gerits@gmail.com");
        Author wim = new Author("Wim Van Leuven", "A brilliant architect interested in BigData, Cloud, development principles and other geeky things", "wim.vanleuven@highestpoint.biz");

        Book myBook = new Book();
        myBook.setTitle("My Book");
        myBook.setLanguage("en");
        myBook.setPages(513);
        myBook.setAuthors(Arrays.asList(daan, wim));

        // -- create our entities
        if (!lilyService.exists(daan)) {
            lilyService.create(daan);
            LOGGER.info("Created " + daan);
        } else {
            LOGGER.info(daan + " already existed");
        }

        if (!lilyService.exists(wim)) {
            lilyService.create(wim);
            LOGGER.info("Created " + wim);
        } else {
            LOGGER.info(wim + " already existed");
        }

        if (!lilyService.exists(myBook)) {
            lilyService.create(myBook);
            LOGGER.info("Created " + myBook);
        } else {
            LOGGER.info(myBook + " already existed");
        }
    }

    public void demoRetrieval() {
        // -- get the book from lily
        Book book = lilyService.get(Book.class, "My Book");

        // -- print the details of our book
        LOGGER.info("Retrieved " + book);
    }

    public void demoResolve() {
        // -- get the record to resolve
        Book book = lilyService.get(Book.class, "My Book");

        // -- resolve additional information
        lilyService.resolve(book, "authors");

        // -- print the details of our book
        LOGGER.info("Resolved " + book);
    }

    public void demoUpdate() {
        // -- get the record to update
        Book book = lilyService.get(Book.class, "My Book");

        // -- change the book data
        book.setLanguage("nl");
        book.setPages(130);

        // -- update the details
        lilyService.update(book);

        // -- print the details of our book
        LOGGER.info("Updated " + book);
    }

    public void demoDelete() {
        // -- get the record to delete
        Book book = lilyService.get(Book.class, "My Book");

        // -- remove the book
        lilyService.delete(book);

        // -- try to get the book, this should result in an exception
        try {
            lilyService.get(Book.class, "My Book");

            // -- it would be incorrect if we got here
            LOGGER.error("Record was not deleted!");
        } catch (LilyException e) {
            LOGGER.info("Record deleted");
        }
    }

}
